import torch
from transformers import AutoTokenizer, AutoModelForSequenceClassification, pipeline
import os

def get_model_size(model):
    """모델의 메모리 크기를 MB 단위로 계산합니다."""
    param_size = 0
    for param in model.parameters():
        param_size += param.nelement() * param.element_size()
    buffer_size = 0
    for buffer in model.buffers():
        buffer_size += buffer.nelement() * buffer.element_size()
    size_all_mb = (param_size + buffer_size) / 1024**2
    return size_all_mb

def main():
    """
    Hugging Face의 사전 학습된 BERT 모델을 로드하고, 8-bit 양자화를 적용한 후
    원본 모델과 양자화된 모델의 크기를 비교하고 추론 결과를 확인하는 스크립트입니다.
    """
    # 사용할 모델과 디바이스 설정
    model_id = "bert-base-uncased"
    device = "cuda" if torch.cuda.is_available() else "cpu"
    
    print(f"사용할 모델: {model_id}")
    print(f"사용할 디바이스: {device}")
    
    if device == "cpu":
        print("\n경고: CUDA를 사용할 수 없어 CPU에서 실행합니다. bitsandbytes 양자화는 CUDA GPU에서만 효과가 있습니다.")
        print("이 스크립트는 CUDA 환경에서 실행하는 것을 권장합니다.")

    # --- 1. 원본 FP32 모델 로드 및 크기 확인 ---
    print("\n--- 1. 원본 FP32 모델 로드 중... ---")
    tokenizer = AutoTokenizer.from_pretrained(model_id)
    fp32_model = AutoModelForSequenceClassification.from_pretrained(model_id)
    
    # 모델을 디바이스로 이동
    fp32_model.to(device)
    
    fp32_model_size = get_model_size(fp32_model)
    print(f"FP32 모델 크기: {fp32_model_size:.2f} MB")

    # --- 2. 8-bit 양자화 모델 로드 및 크기 확인 ---
    # `load_in_8bit=True` 옵션을 사용하여 모델을 8-bit로 양자화합니다.
    # `device_map="auto"`는 모델 레이어를 사용 가능한 디바이스(GPU, CPU, RAM)에 자동으로 분배합니다.
    print("\n--- 2. 8-bit 양자화 모델 로드 중... ---")
    try:
        int8_model = AutoModelForSequenceClassification.from_pretrained(
            model_id,
            load_in_8bit=True,
            device_map="auto"
        )
        int8_model_size = get_model_size(int8_model)
        print(f"INT8 모델 크기: {int8_model_size:.2f} MB")

        size_reduction = 100 * (1 - int8_model_size / fp32_model_size)
        print(f"\n메모리 사용량 감소율: {size_reduction:.2f}%")

    except Exception as e:
        print(f"8-bit 모델 로드 중 오류 발생: {e}")
        print("bitsandbytes가 올바르게 설치되었는지, CUDA 환경이 맞는지 확인하세요.")
        int8_model = None

    # --- 3. 추론 파이프라인으로 결과 비교 ---
    print("\n--- 3. 추론 결과 비교 ---")
    text = "This is a great movie, I really enjoyed it!"
    print(f"입력 텍스트: '{text}'")

    # FP32 모델 추론
    print("\nFP32 모델 추론 결과:")
    classifier_fp32 = pipeline("text-classification", model=fp32_model, tokenizer=tokenizer, device=0 if device=="cuda" else -1)
    result_fp32 = classifier_fp32(text)
    print(result_fp32)

    # INT8 모델 추론
    if int8_model:
        print("\nINT8 모델 추론 결과:")
        # 양자화된 모델은 device_map을 통해 이미 GPU에 할당되어 있으므로 device=-1로 설정해도 GPU에서 실행됩니다.
        classifier_int8 = pipeline("text-classification", model=int8_model, tokenizer=tokenizer)
        result_int8 = classifier_int8(text)
        print(result_int8)
    
    print("\n실습 완료!")
    print("양자화를 통해 모델 크기가 크게 줄어들면서도 추론 결과는 유사하게 유지되는 것을 확인할 수 있습니다.")


if __name__ == "__main__":
    main() 